package jdk.internal.org.objectweb.asm.commons;

import jdk.internal.org.objectweb.asm.Opcodes;
import jdk.internal.org.objectweb.asm.AnnotationVisitor;
import jdk.internal.org.objectweb.asm.TypePath;
import jdk.internal.org.objectweb.asm.Label;
import jdk.internal.org.objectweb.asm.Type;
import jdk.internal.org.objectweb.asm.MethodVisitor;

public class LocalVariablesSorter extends MethodVisitor
{
    private static final Type OBJECT_TYPE;
    private int[] mapping;
    private Object[] newLocals;
    protected final int firstLocal;
    protected int nextLocal;
    private boolean changed;
    
    public LocalVariablesSorter(int n, String s, MethodVisitor methodVisitor) {
        this(327680, n, s, methodVisitor);
        if (this.getClass() != LocalVariablesSorter.class) {
            throw new IllegalStateException();
        }
        return;
    }
    
    protected LocalVariablesSorter(int api, int n, String methodDescriptor, MethodVisitor methodVisitor) {
        super(api, methodVisitor);
        this.mapping = new int[40];
        this.newLocals = new Object[20];
        Type[] argumentTypes = Type.getArgumentTypes(methodDescriptor);
        this.nextLocal = ((0x8 & n) == 0x0);
        for (int i = 0; i < argumentTypes.length; ++i) {
            this.nextLocal += argumentTypes[i].getSize();
        }
        this.firstLocal = this.nextLocal;
    }
    
    @Override
    public void visitVarInsn(int opcode, int n) {
        Type type;
        switch (opcode) {
            case 22:
            case 55: {
                type = Type.LONG_TYPE;
                break;
            }
            case 24:
            case 57: {
                type = Type.DOUBLE_TYPE;
                break;
            }
            case 23:
            case 56: {
                type = Type.FLOAT_TYPE;
                break;
            }
            case 21:
            case 54: {
                type = Type.INT_TYPE;
                break;
            }
            default: {
                type = LocalVariablesSorter.OBJECT_TYPE;
                break;
            }
        }
        this.mv.visitVarInsn(opcode, this.remap(n, type));
    }
    
    @Override
    public void visitIincInsn(int n, int increment) {
        this.mv.visitIincInsn(this.remap(n, Type.INT_TYPE), increment);
    }
    
    @Override
    public void visitMaxs(int maxStack, int n) {
        this.mv.visitMaxs(maxStack, this.nextLocal);
    }
    
    @Override
    public void visitLocalVariable(String name, String s, String signature, Label start, Label end, int n) {
        this.mv.visitLocalVariable(name, s, signature, start, end, this.remap(n, Type.getType(s)));
    }
    
    @Override
    public AnnotationVisitor visitLocalVariableAnnotation(int typeRef, TypePath typePath, Label[] start, Label[] end, int[] array, String s, boolean visible) {
        Type type = Type.getType(s);
        int[] index = new int[array.length];
        for (int i = 0; i < index.length; ++i) {
            index[i] = this.remap(array[i], type);
        }
        return this.mv.visitLocalVariableAnnotation(typeRef, typePath, start, end, index, s, visible);
    }
    
    @Override
    public void visitFrame(int n, int numLocal, Object[] local, int n2, Object[] array) {
        if (n != -1) {
            throw new IllegalStateException("ClassReader.accept() should be called with EXPAND_FRAMES flag");
        }
        if (!this.changed) {
            this.mv.visitFrame(n, numLocal, local, n2, array);
            return;
        }
        Object[] newLocals = new Object[this.newLocals.length];
        System.arraycopy(this.newLocals, 0, newLocals, 0, newLocals.length);
        this.updateNewLocals(this.newLocals);
        int n4;
        for (int n3 = 0, i = 0; i < numLocal; n3 += n4, ++i) {
            Object o = local[i];
            n4 = ((o == Opcodes.LONG || o == Opcodes.DOUBLE) ? 2 : 1);
            if (o != Opcodes.TOP) {
                Type type = LocalVariablesSorter.OBJECT_TYPE;
                if (o == Opcodes.INTEGER) {
                    type = Type.INT_TYPE;
                }
                else if (o == Opcodes.FLOAT) {
                    type = Type.FLOAT_TYPE;
                }
                else if (o == Opcodes.LONG) {
                    type = Type.LONG_TYPE;
                }
                else if (o == Opcodes.DOUBLE) {
                    type = Type.DOUBLE_TYPE;
                }
                else if ((o instanceof String)) {
                    type = Type.getObjectType((String)o);
                }
                this.setFrameLocal(this.remap(n3, type), o);
            }
        }
        int j = 0;
        int numLocal2 = 0;
        for (int n5 = 0; j < this.newLocals.length; ++n5) {
            Object o2 = this.newLocals[j++];
            if (o2 != null && o2 != Opcodes.TOP) {
                this.newLocals[n5] = o2;
                numLocal2 = n5 + 1;
                if (o2 == Opcodes.LONG || o2 == Opcodes.DOUBLE) {
                    ++j;
                }
            }
            else {
                this.newLocals[n5] = Opcodes.TOP;
            }
        }
        this.mv.visitFrame(n, numLocal2, this.newLocals, n2, array);
        this.newLocals = newLocals;
        return;
    }
    
    public int newLocal(Type type) {
        Integer n;
        switch (type.getSort()) {
            case 1:
            case 2:
            case 3:
            case 4:
            case 5: {
                n = Opcodes.INTEGER;
                break;
            }
            case 6: {
                n = Opcodes.FLOAT;
                break;
            }
            case 7: {
                n = Opcodes.LONG;
                break;
            }
            case 8: {
                n = Opcodes.DOUBLE;
                break;
            }
            case 9: {
                n = type.getDescriptor();
                break;
            }
            default: {
                n = type.getInternalName();
                break;
            }
        }
        int localMapping = this.newLocalMapping(type);
        this.setLocalType(localMapping, type);
        this.setFrameLocal(localMapping, n);
        this.changed = true;
        return localMapping;
    }
    
    protected void updateNewLocals(Object[] array) {
    }
    
    protected void setLocalType(int n, Type type) {
    }
    
    private void setFrameLocal(int n, Object o) {
        int length = this.newLocals.length;
        if (n >= length) {
            Object[] newLocals = new Object[Math.max(2 * length, n + 1)];
            System.arraycopy(this.newLocals, 0, newLocals, 0, length);
            this.newLocals = newLocals;
        }
        this.newLocals[n] = o;
    }
    
    private int remap(int n, Type type) {
        if (n + type.getSize() <= this.firstLocal) {
            return n;
        }
        int n2 = 2 * n + type.getSize() - 1;
        int length = this.mapping.length;
        if (n2 >= length) {
            int[] mapping = new int[Math.max(2 * length, n2 + 1)];
            System.arraycopy(this.mapping, 0, mapping, 0, length);
            this.mapping = mapping;
        }
        int localMapping = this.mapping[n2];
        if (localMapping == 0) {
            localMapping = this.newLocalMapping(type);
            this.setLocalType(localMapping, type);
            this.mapping[n2] = localMapping + 1;
        }
        else {
            --localMapping;
        }
        if (localMapping != n) {
            this.changed = true;
        }
        return localMapping;
    }
    
    protected int newLocalMapping(Type type) {
        int nextLocal = this.nextLocal;
        this.nextLocal += type.getSize();
        return nextLocal;
    }
    
    static {
        OBJECT_TYPE = Type.getObjectType("java/lang/Object");
    }
}
